from __future__ import annotations
from functools import lru_cache, reduce
from typing import TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from ._api import Backend, AnyArray

# Backend-independent implementation of low-pass/high-pass filters.


def lowpass_filter_ft(
    backend: Backend,
    img: AnyArray[np.float32],
    cutoff: float,
    order: int = 2,
):
    """Apply a low-pass filter and return the result in Fourier space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return backend.fftn(img)
    weight = nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=False,
        backend=backend,
    )
    return weight * backend.fftn(img)


def lowpass_filter(
    backend: Backend,
    img: AnyArray[np.float32],
    cutoff: float,
    order: int = 2,
):
    """Apply a low-pass filter and return the result in real space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return img
    weight = nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=True,
        backend=backend,
    )
    out = backend.irfftn(weight * backend.rfftn(img))
    return out.real


def highpass_filter_ft(
    backend: Backend,
    img: AnyArray[np.float32],
    cutoff: float,
    order: int = 2,
):
    """Apply a high-pass filter and return the result in Fourier space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return backend.zeros(img.shape, img.dtype)
    weight = 1 - nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=False,
    )
    return weight * backend.fftn(img)


def highpass_filter(
    backend: Backend,
    img: AnyArray[np.float32],
    cutoff: float,
    order: int = 2,
):
    """Apply a high-pass filter and return the result in real space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return backend.zeros(img.shape, img.dtype)
    weight = 1 - nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=True,
    )
    out = backend.irfftn(weight * backend.rfftn(img))
    return out.real


# Modified from skimage.filters._fft_based
@lru_cache(maxsize=4)
def nd_butterworth_weight(
    shape: tuple[int, ...],
    cutoff: float,
    order: int,
    real: bool,
    backend: Backend,
) -> AnyArray[np.complex64]:
    ranges = []
    for d in shape:
        axis = backend.arange(
            -(d - 1) // 2, (d - 1) // 2 + 1, dtype=np.float32
        ) / (d * cutoff)
        ranges.append(backend.ifftshift(axis**2))
    if real:
        limit = shape[-1] // 2 + 1
        ranges[-1] = ranges[-1][:limit]
    q2 = reduce(backend._xp_.add, backend.meshgrid(*ranges, indexing="ij", sparse=True))
    wfilt = 1 / (1 + q2**order)
    return wfilt



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterator,
    NamedTuple,
    cast,
)
import polars as pl

if TYPE_CHECKING:
    from polars.dataframe.groupby import GroupBy
    from .core import Molecules


class MoleculeCutGroup:
    """A groupby-like object for molecules created by :meth:`cutby`."""

    def __init__(self, group: GroupBy[pl.DataFrame], label: str):
        self._group = group
        self._label = label

    def __iter__(self) -> Iterator[tuple[CutEdges, Molecules]]:
        from .core import Molecules

        for key, df in self._group:
            key = cast(str, key)
            gt, le = map(float, key[1:-1].split(", "))
            mole = Molecules.from_dataframe(df.drop(self._label))
            yield CutEdges(gt, le), mole


class CutEdges(NamedTuple):
    gt: float
    le: float

    def __repr__(self) -> str:
        return f"({self.gt}, {self.le}]"



from ._base import (
    BaseAlignmentModel,
    AlignmentResult,
    RotationImplemented,
    TomographyInput,
    AlignmentFactory,
)
from ._concrete import PCCAlignment, ZNCCAlignment

__all__ = [
    "AlignmentResult",
    "AlignmentFactory",
    "PCCAlignment",
    "ZNCCAlignment",
    "BaseAlignmentModel",
    "RotationImplemented",
    "TomographyInput",
]



# pyright: reportPrivateImportUsage=false

from __future__ import annotations
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import (
    Callable,
    Iterable,
    NamedTuple,
    Sequence,
    Union,
)
from typing_extensions import Self

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation

from acryo._rotation import normalize_rotations
from acryo._types import Ranges, subpixel, degree
from acryo._utils import compose_matrices
from acryo._dask import DaskTaskPool, compute
from acryo.backend import Backend, AnyArray, NUMPY_BACKEND
from acryo.alignment._bound import ParametrizedModel


TemplateType = Union[NDArray[np.float32], Sequence[NDArray[np.float32]]]
MaskType = Union[
    NDArray[np.float32],
    Callable[[NDArray[np.float32]], NDArray[np.float32]],
    None,
]
AlignmentFactory = Callable[[TemplateType, MaskType], "BaseAlignmentModel"]
_Template = AnyArray[np.complex64]
_Mask = AnyArray[np.float32]


class AlignmentResult(NamedTuple):
    """The optimal alignment result."""

    label: int
    shift: NDArray[np.float32]
    quat: NDArray[np.float32]
    score: float

    def affine_matrix(self, shape: tuple[int, int, int]) -> NDArray[np.float32]:
        """Return the affine matrix."""
        rotator = Rotation.from_quat(self.quat)
        shift_matrix = np.eye(4, dtype=np.float32)
        shift_matrix[:3, 3] = self.shift
        rot_matrix = compose_matrices(np.array(shape) / 2 - 0.5, [rotator])[0]
        return shift_matrix @ rot_matrix


class TemplateMaskCache:
    _dict: dict[Backend, tuple[_Template, _Mask]]

    def __init__(self):
        self._dict = {}

    def get(self, backend: Backend) -> tuple[_Template, _Mask] | None:
        if out := self._dict.get(backend):
            return out
        if val := next(iter(self._dict.values()), None):
            self._dict[backend] = out = backend.asarray(val[0]), backend.asarray(val[1])
            return out
        return None

    def set(self, backend: Backend, template: _Template, mask: _Mask):
        self._dict[backend] = template, mask


class BaseAlignmentModel(ABC):
    """
    The base class to implement alignment method.

    This class supports subvolume masking, pre-transformation of subvolumes and
    template, optimization of spatial transformation.

    subvolume    template
        |            |
        v            v
       ( soft-masking )
        |            |
        v            v
      (pre-transformation)
        |            |
        +-----+------+
              |
              v
         (alignment)

    Abstract Methods
    ----------------
    >>> def optimize(self, template, reference, max_shifts, quaternion, pos, backend):
    >>>     ...
    >>> def pre_transform(self, image, backend):
    >>>     ...

    """

    _DUMMY_POS = np.array([0.0, 0.0, 0.0], dtype=np.float32)
    _DUMMY_QUAT = np.array([0.0, 0.0, 0.0, 1.0], dtype=np.float32)

    def __init__(
        self,
        template: TemplateType,
        mask: MaskType = None,
    ):
        if isinstance(template, np.ndarray):
            if template.dtype != np.float32:
                template = template.astype(np.float32)
            self._template = template
            self._n_templates = 1
            self._ndim = template.ndim
        else:
            self._template = np.stack(template, axis=0)
            if self._template.dtype != np.float32:
                self._template: NDArray[np.float32] = self._template.astype(np.float32)
            self._n_templates = self._template.shape[0]
            self._ndim = self._template.ndim - 1

        if callable(mask):
            if self._n_templates != 1:
                raise ValueError("Cannot create a mask using multiple templates")
            self._mask: NDArray[np.float32] = mask(self._template)
            if self._template.shape != self._mask.shape:
                raise ValueError(
                    "Shape mismatch in between template image "
                    f"{self._template.shape} and mask image {self._mask.shape})."
                )
        elif mask is None:
            self._mask = np.ones(self._template.shape[-self._ndim :], dtype=np.float32)
        else:
            if self._template.shape[-self._ndim :] != mask.shape:
                raise ValueError(
                    "Shape mismatch in between template image "
                    f"{self._template.shape[-self._ndim :]} and mask image {mask.shape})."
                )
            if mask.dtype not in (np.float32, np.bool_):
                mask = mask.astype(np.float32)
            self._mask = mask

        self._template_mask_cache = TemplateMaskCache()
        self._get_template_and_mask_input(Backend())  # cache the template and mask

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(shape={self._template.shape})"

    @property
    def template(self) -> NDArray[np.float32]:
        """The template image."""
        return self._template

    @property
    def mask(self) -> NDArray[np.float32]:
        """The mask image."""
        return self._mask

    @property
    def input_shape(self) -> tuple[int, ...]:
        """Return the array shape of input images and template."""
        return self._template.shape[-self._ndim :]

    @property
    def has_rotation(self) -> bool:
        """If the alignment model has rotation optimization."""
        return False

    @classmethod
    def with_params(
        cls,
        **params,
    ):
        """Create a BaseAlignmentModel instance with parameters."""
        return ParametrizedModel(cls, **params)

    @abstractmethod
    def _optimize(
        self,
        subvolume: AnyArray[np.complex64],
        template: _Template,
        max_shifts: tuple[float, ...],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], float]:
        """
        Optimization of shift and rotation of subvolume.

        This method uses a subvolume and a template image to find the optimal
        shift/rotation to fit subvolume to template under the restriction of
        maximum shifts.

        Parameters
        ----------
        subvolume : np.ndarray
            Input subvolume. This array has the same dimensionality as the template.
        template : np.ndarray
            Template image.
        max_shifts : tuple of float
            Maximum shifts in each axis. This value should be float pixel. If
            optimization requires integer input, it is better to convert this
            parameter to tuple of integers.

        Returns
        -------
        (3,) np.ndarray, (4,) np.ndarray and float
            Local shift, local rotation and score. "Local" means you don't have
            to consider the position and orientation of molecules.
            - shift ... Shift in pixel.
            - rotation ... Rotation in quaternion. If this cannot be optimized by the
              implemented algorithm, this value can be ``[0., 0., 0., 0.]``.
            - score ... This value Must be a float and larger value should represent
              better results.
        """

    @abstractmethod
    def pre_transform(
        self,
        image: AnyArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.complex64]:
        """Pre-transformation applied to input images (including template)."""

    def _landscape(
        self,
        subvolume: AnyArray[np.complex64],
        template: _Template,
        max_shifts: tuple[float, ...],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32]:
        """Return the landscape of the subvolume."""
        raise NotImplementedError(
            f"_landscape method is not implemented for {type(self).__name__}."
        )

    def _get_template_and_mask_input(
        self,
        backend: Backend | None = None,
    ) -> tuple[_Template, AnyArray[np.float32]]:
        """
        Returns proper template and mask images for alignment.

        Template dimensionality will be dispatched according to the input
        parameters. Returned template should be used in line of the
        :func:`get_alignment_function`.

        Returns
        -------
        array
            Template image(s) and mask image(s). Its axes varies depending on the input.

            - single template image ... 3D
            - many template images ... 4D
        """
        xp = backend or Backend()
        if out := self._template_mask_cache.get(xp):
            return out
        mask = xp.asarray(self._mask)
        if self._n_templates > 1:
            template_input = xp.stack(
                [
                    self.pre_transform(xp.asarray(tmp) * mask, xp)
                    for tmp in self._template
                ],
                axis=0,
            )
        else:
            template_masked = xp.asarray(self._template * self._mask)
            template_input = self.pre_transform(template_masked, xp)
        self._template_mask_cache.set(xp, template_input, mask)
        return template_input, mask

    def align(
        self,
        img: NDArray[np.float32] | AnyArray[np.float32],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32] | None = None,
        pos: NDArray[np.float32] | None = None,
        backend: Backend | None = None,
    ) -> AlignmentResult:
        """
        Align an image using current alignment parameters.

        Parameters
        ----------
        img : np.ndarray
            Subvolume to be aligned
        max_shifts : tuple[float, float, float]
            Maximum shifts along z, y, x axis in pixel.

        Returns
        -------
        AlignmentResult
            Result of alignment.
        """
        xp = backend or Backend()
        if quaternion is None:
            _quat = self._DUMMY_QUAT
        else:
            _quat = quaternion
        if pos is None:
            _pos = self._DUMMY_POS
        else:
            _pos = pos

        if self._is_multiple():
            _align_fn = self._optimize_multiple
        else:
            _align_fn = self._optimize_single
        _template, _mask = self._get_template_and_mask_input(backend=xp)
        return _align_fn(
            xp.asarray(img),
            _template,
            _mask,
            max_shifts,
            _quat,
            _pos,
            xp,
        )

    def fit(
        self,
        img: NDArray[np.float32],
        max_shifts: tuple[float, float, float],
        cval: float | None = None,
        backend: Backend | None = None,
    ) -> tuple[NDArray[np.float32], AlignmentResult]:
        """
        Fit image to template based on the alignment model.

        Parameters
        ----------
        img : np.ndarray
            Input image that will be transformed.
        max_shifts : tuple[float, float, float]
            Maximum shifts along z, y, x axis in pixel.
        cval : float, optional
            Constant value for padding.

        Returns
        -------
        np.ndarray, AlignmentResult
            Transformed input image and the alignment result.
        """
        xp = backend or Backend()
        img_input = xp.asarray(img)
        result = self.align(
            img_input, max_shifts=max_shifts, quaternion=None, pos=None, backend=xp
        )
        mtx = result.affine_matrix(img_input.shape)
        _cval = _normalize_cval(cval, img_input, xp)
        img_trans = xp.affine_transform(img_input, mtx, cval=_cval)
        return xp.asnumpy(img_trans), result

    def landscape(
        self,
        img: NDArray[np.float32],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32] | None = None,
        pos: NDArray[np.float32] | None = None,
        upsample: int = 1,
        backend: Backend | None = None,
    ) -> NDArray[np.float32]:
        """
        Calculate correlation landscape of the input image.

        Parameters
        ----------
        img : np.ndarray
            Subvolume for the landscape calculation.
        max_shifts : tuple[float, float, float]
            Maximum shifts along z, y, x axis in pixel.

        Returns
        -------
        np.ndarray
            N (if single template) or N+1 (if multi-template) dimensional array of
            correlation landscape.
        """
        xp = backend or Backend()
        if quaternion is None:
            _quat = self._DUMMY_QUAT
        else:
            _quat = quaternion
        if pos is None:
            _pos = self._DUMMY_POS
        else:
            _pos = pos
        if self._is_multiple():
            fn = self._landscape_multiple
        else:
            fn = self._landscape_single

        _template, _mask = self._get_template_and_mask_input(backend=xp)
        # calculate the landscape
        lds = fn(
            xp.asarray(img),
            _template,
            _mask,
            max_shifts,
            _quat,
            _pos,
            xp,
        )

        if upsample > 1:
            if not self._is_multiple():
                coords = _build_mesh(lds.shape, max_shifts, upsample, xp)
                lds_upsampled = xp.map_coordinates(
                    lds, coords, order=3, mode="constant", cval=0.0, prefilter=True
                )
            else:
                coords = _build_mesh(lds.shape[1:], max_shifts, upsample, xp)
                all_lds = [
                    xp.map_coordinates(
                        l, coords, order=3, mode="constant", cval=0.0, prefilter=True
                    )
                    for l in lds
                ]
                lds_upsampled = xp.stack(all_lds, axis=0)
            return xp.asnumpy(lds_upsampled)
        return xp.asnumpy(lds)

    def _landscape_single(
        self,
        subvolume: AnyArray[np.float32],
        template: _Template,
        mask: AnyArray[np.float32],
        max_shifts: tuple[float, ...],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ):
        return self._landscape(
            self.pre_transform(subvolume * mask, backend),
            template,
            max_shifts=max_shifts,
            quaternion=quaternion,
            pos=pos,
            backend=backend,
        )

    def _landscape_multiple(
        self,
        subvolume: AnyArray[np.float32],
        template_list: _Template,
        mask_list: Iterable[AnyArray[np.float32]],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32]:
        out: list[AnyArray[np.float32]] = []
        for template, mask in zip(template_list, mask_list):
            lnd = self._landscape(
                self.pre_transform(subvolume * mask, backend),
                template,
                max_shifts=max_shifts,
                quaternion=quaternion,
                pos=pos,
                backend=backend,
            )
            out.append(lnd)
        return backend.stack(out, axis=0)

    def _optimize_single(
        self,
        subvolume: AnyArray[np.float32],
        template: _Template,
        mask: AnyArray[np.float32],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AlignmentResult:
        out = self._optimize(
            self.pre_transform(subvolume * mask, backend),
            template,
            max_shifts=max_shifts,
            quaternion=quaternion,
            pos=pos,
            backend=backend,
        )
        return AlignmentResult(0, *out)

    def _optimize_multiple(
        self,
        subvolume: AnyArray[np.float32],
        template_list: _Template,
        mask_list: AnyArray[np.float32],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AlignmentResult:
        all_shifts: list[np.ndarray] = []
        all_quat: list[np.ndarray] = []
        all_score: list[float] = []
        for template, mask in zip(template_list, mask_list):
            shift, quat, score = self._optimize(
                self.pre_transform(subvolume * mask, backend),
                template,
                max_shifts=max_shifts,
                quaternion=quaternion,
                pos=pos,
                backend=backend,
            )
            all_shifts.append(shift)
            all_quat.append(quat)
            all_score.append(score)

        iopt = int(np.argmax(all_score))
        return AlignmentResult(iopt, all_shifts[iopt], all_quat[iopt], all_score[iopt])

    def _is_multiple(self) -> bool:
        return self.niter > 1

    @property
    def is_multi_templates(self) -> bool:
        """
        Whether alignment parameters requires multi-templates.
        "Multi-template" includes alignment with subvolume rotation.
        """
        return self._is_multiple()

    @property
    def niter(self) -> int:
        """Number of templates."""
        return self._n_templates


class RotationImplemented(BaseAlignmentModel):
    """
    An alignment model implemented with default rotation optimizer.

    If ``optimize`` does not support rotation optimization, this class implements
    simple parameter searching algorithm to it. Thus, ``optimize`` only has to
    optimize shift of images.
    """

    def __init__(
        self,
        template: TemplateType,
        mask: MaskType = None,
        rotations: Ranges | None = None,
    ):
        self.quaternions = normalize_rotations(rotations)
        self._n_rotations = self.quaternions.shape[0]
        super().__init__(template=template, mask=mask)

    @classmethod
    def with_params(
        cls,
        *,
        rotations: Ranges | None = None,
    ) -> ParametrizedModel[Self]:
        """Create an alignment model instance with parameters."""
        return ParametrizedModel(cls, rotations=rotations)

    @property
    def has_rotation(self) -> bool:
        """If the alignment model has rotation optimization."""
        return self._n_rotations > 1

    def align(
        self,
        img: NDArray[np.float32],
        max_shifts: tuple[subpixel, subpixel, subpixel],
        quaternion: NDArray[np.float32] | None,
        pos: NDArray[np.float32] | None,
        backend: Backend | None = None,
    ) -> AlignmentResult:
        """
        Align an image using current alignment parameters.

        Parameters
        ----------
        img : np.ndarray
            Subvolume to be aligned
        max_shifts : tuple of float
            Maximum shifts along z, y, x axis in pixel.

        Returns
        -------
        AlignmentResult
            Result of alignment.
        """
        iopt, shift, _, corr = super().align(img, max_shifts, quaternion, pos, backend)
        quat = self.quaternions[iopt % self._n_rotations]
        return AlignmentResult(label=iopt, shift=shift, quat=quat, score=corr)

    def fit(
        self,
        img: NDArray[np.float32],
        max_shifts: tuple[subpixel, subpixel, subpixel],
        cval: float | None = None,
        backend: Backend | None = None,
    ) -> tuple[NDArray[np.float32], AlignmentResult]:
        """
        Fit image to template based on the alignment model.

        Unlike ``BaseAlignmentModel``, rotation optimization is executed in
        parallel to boost calculation.

        Parameters
        ----------
        img : np.ndarray
            Input image that will be transformed.
        max_shifts : tuple of float
            Maximum shifts along z, y, x axis in pixel.
        cval : float, optional
            Constant value for padding.

        Returns
        -------
        np.ndarray, AlignmentResult
            Transformed input image and the alignment result.
        """
        xp = backend or Backend()
        pool = DaskTaskPool.from_func(self._optimize)
        pos = np.zeros(3, dtype=np.float32)
        img_input = xp.asarray(img)
        _template, _mask = self._get_template_and_mask_input(backend=xp)
        for quat, tmp, mask in zip(self.quaternions, _template, _mask):
            pool.add_task(
                self.pre_transform(img_input * mask, xp),
                tmp,
                max_shifts,
                quat,
                pos=pos,
                backend=xp,
            )
        results = pool.compute()
        scores = [x[2] for x in results]
        iopt = np.argmax(scores)
        opt_result = results[iopt]
        result = AlignmentResult(
            label=0,
            shift=opt_result[0],
            quat=self.quaternions[iopt],
            score=opt_result[2],
        )

        mtx = result.affine_matrix(img_input.shape)
        _img_cval = _normalize_cval(cval, img_input, xp)
        img_trans = xp.affine_transform(img_input, mtx, cval=_img_cval)
        return xp.asnumpy(img_trans), result

    def _transform_template(
        self,
        temp: AnyArray[np.float32],
        matrix: NDArray[np.float32],
        cval: float | None = None,
        order: int = 3,
        prefilter: bool = True,
        backend: Backend = NUMPY_BACKEND,
    ) -> _Template:
        _cval = _normalize_cval(cval, temp, backend)
        temp_transformed = backend.affine_transform(
            temp, matrix=matrix, cval=_cval, order=order, prefilter=prefilter
        )
        return self.pre_transform(temp_transformed, backend)

    def _get_template_and_mask_input(
        self,
        backend: Backend | None = None,
    ) -> tuple[_Template, NDArray[np.float32]]:
        """
        Returns proper template image for alignment.

        Template dimensionality will be dispatched according to the input
        parameters. Returned template should be used in line of the
        :func:`get_alignment_function`.

        Returns
        -------
        np.ndarray
            Template image(s). Its axes varies depending on the input.

            - no rotation, single template image ... 3D
            - has rotation, single template image ... 4D
            - no rotation, many template images ... 4D
            - has rotation, many template images ... 4D and when iterated over
              the first axis yielded images will be (rot0, temp0),
              (rot0, temp1), ...
        """
        xp = backend or Backend()
        if self._n_rotations > 1:
            rotators = [Rotation.from_quat(r).inv() for r in self.quaternions]
            matrices = compose_matrices(
                np.array(self._template.shape[-3:]) / 2 - 0.5, rotators
            )
            cval = float(np.percentile(self._template, 1))
            if self._n_templates > 1:
                inputs_templates = [
                    xp.spline_filter(
                        tmp * self._mask,
                        order=3,
                        mode="constant",
                        output=np.float32,
                    )
                    for tmp in self._template
                ]
                pool_template = DaskTaskPool.from_func(self._transform_template)
                pool_mask = DaskTaskPool.from_func(xp.affine_transform)
                ntmp = len(inputs_templates)
                for mat in matrices:
                    for tmp in inputs_templates:
                        pool_template.add_task(
                            tmp,
                            mat,
                            order=3,
                            cval=cval,
                            prefilter=False,
                            backend=xp,
                        )
                    pool_mask.add_tasks(
                        ntmp, self._mask, mat, order=3, mode="nearest", prefilter=False
                    )
            else:
                template_masked = xp.spline_filter(
                    self._template * self._mask,
                    order=3,
                    output=np.float32,
                    mode="constant",
                )
                pool_template = DaskTaskPool.from_func(self._transform_template)
                pool_mask = DaskTaskPool.from_func(xp.affine_transform)
                for mat in matrices:
                    pool_template.add_task(
                        template_masked,
                        mat,
                        order=3,
                        cval=cval,
                        prefilter=False,
                        backend=xp,
                    )
                    pool_mask.add_task(
                        self._mask, mat, order=3, mode="nearest", prefilter=False
                    )

            _templates, _masks = compute(
                (
                    pool_template.asarrays(self.input_shape, dtype=np.complex64),
                    pool_mask.asarrays(self.input_shape, dtype=np.float32),
                )
            )
            template_input = xp.stack(_templates, axis=0)
            mask_input = xp.stack(_masks, axis=0)
        else:
            pool = DaskTaskPool.from_func(self.pre_transform)
            if self._n_templates > 1:
                for tmp in self._template:
                    pool.add_task(tmp * self._mask, xp)
                template_input = xp.stack(pool.compute(), axis=0)
                mask_input = xp.stack(
                    [xp.asarray(self._mask)] * self._n_templates, axis=0
                )

            else:
                # NOTE: dask.compute is always called once inside this method.
                template_masked = xp.asarray(self._template * self._mask)
                template_input = pool.add_task(template_masked, xp).compute()[0]
                mask_input = xp.asarray(self._mask)

        self._template_mask_cache.set(xp, template_input, mask_input)
        return template_input, mask_input

    def _is_multiple(self) -> bool:
        return self._n_templates * self._n_rotations > 1

    @property
    def niter(self) -> int:
        """Number of iteration per sub-volume."""
        return self._n_templates * self._n_rotations


class TomographyInput(RotationImplemented):
    """
    An alignment model that implements missing-wedge masking and low-pass filter.

    This alignment model is useful for subtomogram averaging of real experimental
    data with limited tilt ranges. Template image will be masked with synthetic
    missing-wedge mask in the frequency domain.
    """

    def __init__(
        self,
        template: TemplateType,
        mask: MaskType = None,
        rotations: Ranges | None = None,
        cutoff: float | None = None,
        tilt_range: tuple[degree, degree] | None = None,
    ):
        self._cutoff = cutoff or 1.0
        if tilt_range is not None:
            deg0, deg1 = tilt_range
            if deg0 >= deg1:
                raise ValueError("Tilt range must be in form of (min, max).")
        self._tilt_range = tilt_range
        super().__init__(template, mask, rotations)

    @classmethod
    def with_params(
        cls,
        *,
        rotations: Ranges | None = None,
        cutoff: float | None = None,
        tilt_range: tuple[degree, degree] | None = None,
    ) -> ParametrizedModel[Self]:
        """Create an alignment model instance with parameters."""
        return ParametrizedModel(
            cls,
            rotations=rotations,
            cutoff=cutoff,
            tilt_range=tilt_range,
        )

    def pre_transform(
        self, image: AnyArray[np.float32], backend: Backend
    ) -> AnyArray[np.complex64]:
        """Apply low-pass filter without IFFT."""
        return backend.lowpass_filter_ft(image, cutoff=self._cutoff)

    def masked_difference(
        self,
        image: NDArray[np.float32],
        quaternion: NDArray[np.float32],
        backend: Backend | None = None,
    ) -> NDArray[np.float32]:
        """
        Difference between an image and the template, considering the missing wedge.

        Parameters
        ----------
        image : 3D array
            Input image, usually a subvolume from a tomogram.
        quaternion : (4,) array
            Rotation of the image, usually the quaternion array of a Molecules
            object.

        Returns
        -------
        3D array
            Difference map.
        """
        if self._n_templates > 1:
            raise NotImplementedError(
                "Masked difference is not implemented for multi-template."
            )
        xp = backend or Backend()
        _template, _mask = self._get_template_and_mask_input(xp)
        image_input = self.pre_transform(xp.asarray(image) * _mask, xp)
        mw = self._get_missing_wedge_mask(quaternion, xp)
        template_masked = xp.ifftn(_template * mw).real
        img_input = xp.ifftn(image_input * mw).real
        return xp.asnumpy(img_input - template_masked)

    def mask_missing_wedge(
        self,
        image: NDArray[np.complex64],
        quaternion: NDArray[np.float32],
        backend: Backend | None = None,
    ) -> NDArray[np.complex64]:
        """Apply missing wedge mask in the frequency domain."""
        xp = backend or Backend()
        mask = self._get_missing_wedge_mask(quaternion, xp)
        return xp.asnumpy(xp.asarray(image) * mask)

    def get_missing_wedge_mask(
        self, quat: NDArray[np.float32], backend: Backend | None = None
    ):
        """
        Create a mask that covers tomographical missing wedge.

        Parameters
        ----------
        quat : (4,) array
            Quaternion representation of the orientation of the subvolume.

        Returns
        -------
        np.ndarray or 1
            Missing wedge mask array.
        """
        xp = backend or Backend()
        return self._get_missing_wedge_mask(quat, xp)

    def _get_missing_wedge_mask(
        self,
        quat: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32] | int:
        """Create a mask that covers tomographical missing wedge."""
        if self._tilt_range is None:
            return 1
        return backend.missing_wedge_mask(
            rotator=Rotation.from_quat(quat),
            tilt_range=self._tilt_range,
            shape=self.input_shape,
        )


def _normalize_cval(
    cval: float | None, img: AnyArray[np.float32], backend: Backend
) -> float:
    if cval is None:
        _cval = float(backend.percentile(img, 1))
    else:
        _cval = cval
    return _cval


@lru_cache(maxsize=2)
def _build_mesh(
    shape: tuple[int, int, int],
    max_shifts: tuple[float, float, float],
    upsample: int,
    backend: Backend,
) -> AnyArray[np.float32]:
    upsampled_max_shifts = (np.asarray(max_shifts) * upsample).astype(np.int32)
    center = np.array(shape) / 2 - 0.5
    mesh = backend.meshgrid(
        *[
            backend.arange(-width, width + 1) / upsample + c
            for c, width in zip(center, upsampled_max_shifts)
        ],
        indexing="ij",
    )
    return backend.stack(mesh, axis=0)



# pyright: reportPrivateImportUsage=false
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, SupportsIndex
import numpy as np
from dask import array as da

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from ._dask_pca import DaskPCA as PCA
    from sklearn.cluster import KMeans
    from typing_extensions import Self
    from numpy.typing import NDArray


class PcaClassifier:
    """A PCA (Principal component analysis) and k-means based image classifier."""

    def __init__(
        self,
        image_stack: NDArray[np.float32] | da.Array,
        mask_image: NDArray[np.float32] | None = None,
        n_components: int = 2,
        n_clusters: int = 2,
        seed: int | None = 0,
    ):
        from ._dask_pca import DaskPCA as PCA
        from sklearn.cluster import KMeans

        if mask_image is None:
            self._mask = 1
        else:
            self._mask = mask_image

        if isinstance(image_stack, np.ndarray):
            image_stack = da.from_array(image_stack)
        if not isinstance(image_stack, da.Array):
            raise TypeError("image_stack must be a numpy array or a dask array.")
        self._image = image_stack
        self._n_image = image_stack.shape[0]
        self._shape = image_stack.shape[1:]  # shape of a single image
        self.n_components = n_components
        self.n_clusters = n_clusters

        self._pca = PCA(n_components=n_components)
        self._kmeans = KMeans(n_clusters=n_clusters, random_state=seed, n_init=10)

    @property
    def pca(self) -> PCA:
        """The PCA object."""
        return self._pca

    @property
    def kmeans(self) -> KMeans:
        """The k-means object."""
        return self._kmeans

    @property
    def labels(self) -> NDArray[np.int32] | None:
        return self._labels

    def run(self) -> Self:
        """Run PCA and k-means clustering."""
        _flat_image = self._image_flat(mask=True)
        self._pca.fit(_flat_image)
        self._labels = self._kmeans.fit_predict(self.get_transform())
        return self

    def transform(self, input: da.Array, mask: bool = True) -> NDArray[np.float32]:
        """Transform the input image into the principal component space."""
        if mask:
            input = input * self._mask
        flat = input.reshape(input.shape[0], -1)
        return self._pca.transform(flat).compute()

    def predict(self, input: da.Array) -> NDArray[np.int32]:
        """Predict which labels the input images belong to."""
        transformed = self.transform(input)
        labels = self._kmeans.predict(transformed)
        return labels

    def _image_flat(self, mask: bool = False) -> da.Array:
        if mask:
            _input = self._image * self._mask
        else:
            _input = self._image
        _flat_images = _input.reshape(self._n_image, -1)
        return _flat_images

    def get_transform(self, labels: Iterable[int] | None = None) -> NDArray[np.float32]:
        """
        Get the transformed vectors from the input images.

        Returns
        -------
        da.Array
            Transormed vectors. If input image stack P images, then
            (P, n_components) array will be returned.
        """
        if labels is None:
            flat = self._image_flat(mask=True)
        else:
            if not isinstance(labels, list):
                labels = list(labels)
                if not isinstance(labels[0], SupportsIndex):
                    raise TypeError("labels must be a list of integers.")
            flat = self._image_flat(mask=True)[labels]
        return self._pca.transform(flat).compute()

    def plot_transform(
        self,
        labels: Iterable[int] | None = None,
        bases: tuple[int, int] = (0, 1),
        ax: Axes | None = None,  # type: ignore
    ) -> Axes:
        ax0, ax1 = bases

        transformed = self.get_transform(labels)
        if ax is None:
            import matplotlib.pyplot as plt

            ax: Axes = plt.gca()
        ax.scatter(transformed[:, ax0], transformed[:, ax1])
        return ax

    def get_bases(self) -> np.ndarray:
        """
        Get base images (principal axes) as image stack.

        Returns
        -------
        np.ndarray
            Same axes as input image stack, while the axis "p" corresponds to
            the identifier of bases.
        """
        return self.pca.components_.reshape(self.n_components, *self._shape)  # type: ignore

    def split_clusters(self) -> list[da.Array]:
        """
        Split input image stack into list of image stacks according to the labels.

        This method must be called after k-means clustering is conducted, otherwise
        only one cluster will be returned. If input image stack has ``"pzyx"`` axes,
        list of ``"pzyx"`` images will be returned.
        """
        output: list[da.Array] = []
        for i in range(self.n_clusters):
            img0 = self._image[self._labels == i]
            output.append(img0)
        return output



from __future__ import annotations

from typing import Iterable, Any, Sequence
import numpy as np
from numpy.typing import NDArray
import polars as pl


def get_feature_list(corr_max, local_shifts, rotvec) -> list[pl.Series]:
    features = [
        pl.Series("score", corr_max),
        pl.Series("align-dz", np.round(local_shifts[:, 0], 2)),
        pl.Series("align-dy", np.round(local_shifts[:, 1], 2)),
        pl.Series("align-dx", np.round(local_shifts[:, 2], 2)),
        pl.Series("align-dzrot", np.round(rotvec[:, 0], 5)),
        pl.Series("align-dyrot", np.round(rotvec[:, 1], 5)),
        pl.Series("align-dxrot", np.round(rotvec[:, 2], 5)),
    ]
    return features


def dict_iterrows(d: dict[str, Iterable[Any]]):
    """
    Generater similar to pl.DataFrame.iterrows().

    >>> dict_iterrows({'a': [1, 2, 3], 'b': [4, 5, 6]})

    will yield {'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}.
    """
    keys = d.keys()
    value_iters = [iter(v) for v in d.values()]

    dict_out = dict.fromkeys(keys, None)
    while True:
        try:
            for k, viter in zip(keys, value_iters):
                dict_out[k] = next(viter)
            yield dict_out
        except StopIteration:
            break


def allocate(size: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    # shift in local Cartesian
    local_shifts = np.zeros((size, 3))

    # maximum ZNCC
    corr_max = np.zeros(size)

    # rotation (quaternion) in local Cartesian
    local_rot = np.zeros((size, 4))
    local_rot[:, 3] = 1  # identity map in quaternion

    return local_shifts, local_rot, corr_max


def random_splitter(
    rng: np.random.Generator,
    nmole: int,
) -> tuple[NDArray[np.bool_], NDArray[np.bool_]]:
    sl = rng.choice(np.arange(nmole), nmole // 2).tolist()
    indices0 = np.zeros(nmole, dtype=np.bool_)
    indices0[sl] = True
    indices1 = np.ones(nmole, dtype=np.bool_)
    indices1[sl] = False
    return indices0, indices1


def normalize_shape(a: int | Sequence[int], ndim: int):
    if isinstance(a, int):
        _output_shape = (a,) * ndim
    else:
        _output_shape = tuple(a)
    return _output_shape



from __future__ import annotations

from functools import lru_cache
import math
from typing import TYPE_CHECKING
from scipy.spatial.transform import Rotation
from acryo._types import degree
import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from ._api import AnyArray, Backend


def missing_wedge_mask(
    backend: Backend,
    rotator: Rotation,
    tilt_range: tuple[degree, degree],
    shape: tuple[int, int, int],
) -> AnyArray[np.float32]:
    """
    Create a binary mask that covers tomographical missing wedge.

    Note that the mask is not shifted to the center of the Fourier domain.
    ``np.fft.fftn(img) * mask`` will be the correct way to apply the mask.

    Parameters
    ----------
    rotator : Rotation
        The rotation object that describes the direction of the mask.
    tilt_range : tuple of float
        The range of tilt angles in degrees.
    shape : tuple of int
        The shape of the mask.

    Returns
    -------
    np.ndarray or float
        Missing wedge mask. If ``tilt_range`` is None, 1 will be returned.
    """
    normal0, normal1 = _get_unrotated_normals(tilt_range)
    shape_vector = np.array(shape, dtype=np.float32)
    rotator_inv = rotator.inv()
    normal0 = rotator_inv.apply(normal0 * shape_vector)
    normal1 = rotator_inv.apply(normal1 * shape_vector)
    vectors = _get_indices(shape, backend)
    dot0 = vectors.dot(backend.asarray(normal0))
    dot1 = vectors.dot(backend.asarray(normal1))
    missing = dot0 * dot1 <= 0
    return missing


@lru_cache(maxsize=8)
def _get_unrotated_normals(
    tilt_range: tuple[degree, degree]
) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
    radmin, radmax = np.deg2rad(tilt_range)
    ang0 = np.pi - radmin
    ang1 = np.pi - radmax
    return (
        np.array([np.cos(ang0), 0, np.sin(ang0)], dtype=np.float32),
        np.array([np.cos(ang1), 0, np.sin(ang1)], dtype=np.float32),
    )


@lru_cache(maxsize=32)
def _get_indices(shape: tuple[int, int, int], backend: Backend) -> AnyArray[np.float32]:
    inds = backend.indices(shape, dtype=np.float32)
    for ind, s in zip(inds, shape):
        # Note that the shifts in indices must resemble the shifts in fftshift.
        ind -= math.ceil(s / 2)
    return backend.fftshift(
        backend.stack(list(inds), axis=-1), axes=(0, 1, 2)
    )  # type: ignore



__version__ = "0.4.1"

from acryo.loader import (
    SubtomogramLoader,
    BatchLoader,
    MockLoader,
)
from acryo.molecules import Molecules
from acryo.simulator import TomogramSimulator

imread = SubtomogramLoader.imread

__all__ = [
    "Molecules",
    "SubtomogramLoader",
    "BatchLoader",
    "MockLoader",
    "TomogramSimulator",
    "imread",
]



from __future__ import annotations

from typing import TypeVar
import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.spatial.transform import Rotation

_T = TypeVar("_T", bound=np.number)


def axes_to_rotator(z, y) -> Rotation:
    ref = _normalize(np.atleast_2d(y))

    n = ref.shape[0]
    yx = np.arctan2(ref[:, 2], ref[:, 1])
    zy = np.arctan(-ref[:, 0] / np.abs(ref[:, 1]))

    rot_vec_yx = np.zeros((n, 3))
    rot_vec_yx[:, 0] = yx
    rot_yx = Rotation.from_rotvec(rot_vec_yx)

    rot_vec_zy = np.zeros((n, 3))
    rot_vec_zy[:, 2] = zy
    rot_zy = Rotation.from_rotvec(rot_vec_zy)

    rot1 = rot_yx * rot_zy

    if z is None:
        return rot1

    vec = _normalize(np.atleast_2d(_extract_orthogonal(ref, z)))

    vec_trans = rot1.apply(vec, inverse=True)  # in zx-plane

    thetas = np.arctan2(vec_trans[..., 0], vec_trans[..., 2]) - np.pi / 2

    rot_vec_zx = np.zeros((n, 3))
    rot_vec_zx[:, 1] = thetas
    rot2 = Rotation.from_rotvec(rot_vec_zx)

    return rot1 * rot2


def from_euler_xyz_coords(
    angles: ArrayLike, seq: str = "ZXZ", degrees: bool = False
) -> Rotation:
    """Create a rotator using zyx-coordinate system, from Euler angles."""
    seq = translate_euler(seq)
    angles = np.asarray(angles)
    return Rotation.from_euler(seq, angles[..., ::-1], degrees)


def translate_euler(seq: str) -> str:
    table = str.maketrans({"x": "z", "z": "x", "X": "Z", "Z": "X"})
    return seq[::-1].translate(table)


def _normalize(a: NDArray[_T]) -> NDArray[_T]:
    """Normalize vectors to length 1. Input must be (N, 3)."""
    return a / np.sqrt(np.sum(a**2, axis=1))[:, np.newaxis]


def _extract_orthogonal(a: NDArray[_T], b: NDArray[_T]) -> NDArray[_T]:
    """Extract component of b orthogonal to a."""
    a_norm = _normalize(a)
    return b - np.sum(a_norm * b, axis=1)[:, np.newaxis] * a_norm



from __future__ import annotations

import numpy as np
from scipy import ndimage as ndi
from acryo.pipe._curry import provider_function
from acryo._reader import REG
from acryo._types import nm


@provider_function
def from_file(scale: nm, path: str, original_scale: float | None = None, tol=0.01):
    """
    An image provider function with rescaling.

    This function will provide a subtomogram loader with a resized image from a file.
    Will be used for the template images or the mask images.

    >>> loader.align(
    ...     template=from_file("path/to/template.mrc"),
    ...     mask=from_file("path/to/mask.mrc"),
    ... )

    Parameters
    ----------
    path : str
        Path to the image.
    original_scale : float, optional
        If given, this value will be used as the image scale (nm/pixel) instead
        of the scale extracted from the image metadata.
    tol : float
        Tolerance of the scale difference. If the relative difference is smaller than
        this, the image will not be resized.
    """
    if original_scale is not None and original_scale <= 0:
        raise ValueError("original_scale must be positive")
    img, img_scale = REG.imread_array(path)
    if original_scale is None:
        original_scale = img_scale
    ratio = original_scale / scale
    if abs(ratio - 1) < tol:
        return img
    return ndi.zoom(img, ratio, order=3, prefilter=True, mode="reflect")


@provider_function
def from_gaussian(
    scale: nm,
    shape: tuple[nm, nm, nm],
    sigma: nm | tuple[nm, nm, nm] = 1.0,
    shift: tuple[nm, nm, nm] = (0.0, 0.0, 0.0),
):
    """
    An image provider function by a Gaussian function.

    This function will provide a Gaussian particle with given shape, sigma and shift from
    the center.

    >>> loader.align(
    ...     template=from_gaussian(shape=(4.8, 4.8, 4.8), sigma=1.2),
    ...     mask=from_file("path/to/mask.mrc"),
    ... )

    Parameters
    ----------
    shape : float or tuple of float
        Shape of the output image in nm.
    sigma : float or tuple of float
        Standard deviation of the Gaussian particle in nm.
    shift : tuple of float, optional
        Shift of the Gaussian particle from the center in nm.
    """
    if np.isscalar(sigma):
        sigma = (sigma,) * 3
    if np.isscalar(shape):
        shape = (shape,) * 3
    sigma_px = np.array(sigma) / scale
    shape_subpix = np.array(shape) / scale
    center_subpix = shape_subpix + np.array(shift) / scale
    shape_px = tuple(np.round(shape_subpix).astype(np.int32))

    crds = np.indices(shape_px, dtype=np.float32)

    return np.exp(
        -0.5
        * sum((xx - c) / sg for xx, c, sg in zip(crds, center_subpix, sigma_px)) ** 2
    )


@provider_function
def from_array(
    scale: nm, img: np.ndarray, original_scale: float = 1.0, tol: float = 0.01
):
    """
    An image provider function using existing image array.

    This function will provide a subtomogram loader with a resized image from an array.
    Will be used for the template images or the mask images.

    >>> loader.align(
    ...     template=from_array(img, original_scale=0.28),
    ...     mask=from_file("path/to/mask.mrc"),
    ... )

    Parameters
    ----------
    img : np.ndarray
        Input image array. Must be 3D.
    original_scale : float, optional
        If given, this value will be used as the image scale (nm/pixel) instead
        of the scale extracted from the image metadata.
    tol : float
        Tolerance of the scale difference. If the relative difference is smaller than
        this, the image will not be resized.
    """
    if original_scale is not None and original_scale <= 0:
        raise ValueError("original_scale must be positive")
    if img.ndim != 3:
        raise ValueError("img must be 3D")
    ratio = original_scale / scale
    if abs(ratio - 1) < tol:
        return img
    out = ndi.zoom(img, ratio, order=3, prefilter=True, mode="reflect")
    if not out.dtype == np.float32:
        out = out.astype(np.float32)
    return out


@provider_function
def from_atoms(
    scale: nm,
    atoms: np.ndarray,
    weights: np.ndarray | None = None,
    center: tuple[nm, nm, nm] | None = None,
):
    """
    An image provider function using a point cloud.

    Given an array of atoms, such as data extracted from a PDB file, this function
    can generate a 3D image of the atoms by simply building a histogram.

    Parameters
    ----------
    atoms : (N, 3) array
        Atoms coordinates in nanometer.
    weights : np.ndarray, optional
        weights of the atoms.
    center : tuple of float, optional
        Coordinates of the image center. If not given, the geometric center of the atoms
        will be used.
    """
    if atoms.ndim != 2 or atoms.shape[1] != 3:
        raise ValueError("atoms must be a 2D array with shape (n, 3)")
    if center is None:
        center = np.mean(atoms, axis=0)
    center = np.asarray(center)[np.newaxis]
    coords = (atoms - center) / scale
    rmax = np.max(np.sqrt(np.sum(coords**2, axis=1)))  # the furthest in pixels
    size = int(np.ceil(rmax * 2))
    lims = (-size / 2, size / 2)

    counts, _ = np.histogramdd(
        coords, bins=(size,) * 3, range=(lims,) * 3, weights=weights
    )
    return counts



from __future__ import annotations
import numpy as np


def gaussian_blob(indices, center: tuple[float, float, float], sigma: float):
    zz, yy, xx = indices
    return np.exp(
        -((zz - center[0]) ** 2 + (yy - center[1]) ** 2 + (xx - center[2]) ** 2)
        / sigma ** 2
        / 2
    )


def blobs() -> np.ndarray:
    shape = (40, 40, 40)
    centers = [(17, 17, 17), (17, 17, 26), (24, 17, 18)]
    sigmas = [5.4, 3.6, 2.7]
    img = np.zeros(shape, dtype=np.float32)
    inds = np.indices(shape, dtype=np.float32)
    for center, sigma in zip(centers, sigmas):
        img += gaussian_blob(inds, center, sigma)

    return img


def spiral(radius: float = 4.0, freq: float = 1.0) -> np.ndarray:
    shape = (40, 40, 40)
    img = np.zeros(shape, dtype=np.float32)
    inds = np.indices(shape, dtype=np.float32)
    ys = np.linspace(5, 35, 100)
    zs = radius * np.sin(ys * freq) + (shape[0] - 1) / 2
    xs = radius * np.cos(ys * freq) + (shape[2] - 1) / 2
    for center in np.stack([zs, ys, xs], axis=1):
        img += gaussian_blob(inds, center, 1.0)
    img /= img.max()

    return img



from typing_extensions import TypeAlias

RangeLike = tuple[float, float]
Ranges = tuple[RangeLike, RangeLike, RangeLike]

# type alias
nm: TypeAlias = float
degree: TypeAlias = float
pixel: TypeAlias = int
subpixel: TypeAlias = float



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from scipy import ndimage as ndi
from acryo.pipe._curry import converter_function
from acryo._types import nm


@converter_function
def threshold_otsu(
    img: NDArray[np.float32], scale: nm, bins: int = 256
) -> NDArray[np.bool_]:
    """
    Pipe operation that binarize an image using Otsu's method.

    Parameters
    ----------
    bins : int, default is 256
        Number of bins to build histogram.
    """
    hist, edges = np.histogram(img.ravel(), bins=bins)
    centers: NDArray[np.float32] = (edges[:-1] + edges[1:]) / 2
    npixel0 = np.cumsum(hist)
    npixel1 = img.size - npixel0

    nonzero0 = npixel0 != 0
    nonzero1 = npixel1 != 0

    mean0 = np.zeros_like(centers)
    mean1 = np.zeros_like(centers)
    product = hist * centers
    mean0[nonzero0] = np.cumsum(product)[nonzero0] / npixel0[nonzero0]
    mean1[nonzero1] = (np.cumsum((product)[nonzero1][::-1]) / npixel1[nonzero1][::-1])[
        ::-1
    ]

    s = npixel0 * npixel1 * (mean0 - mean1) ** 2

    imax = np.argmax(s)
    thr = centers[imax]
    return img > thr


@converter_function
def dilation(img: NDArray[np.bool_], scale: nm, radius: nm) -> NDArray[np.bool_]:
    """
    Pipe operation that dilate (or erode) a binary image using a circular structure.

    Parameters
    ----------
    radius : float
        Radius of the structure element in nanometer. If negative, erosion is applied.
    """
    radius_px = abs(radius / scale)
    if radius_px < 1:
        return img
    r = int(np.ceil(radius_px))
    size = 2 * r + 1
    zz, yy, xx = np.indices((size,) * 3)
    structure = (xx - r) ** 2 + (yy - r) ** 2 + (zz - r) ** 2 <= r**2
    if radius < 0:
        out = ndi.binary_erosion(img, structure=structure, border_value=False)
    elif radius > 0:
        out = ndi.binary_dilation(img, structure=structure, border_value=False)
    return out  # type: ignore


@converter_function
def gaussian_smooth(
    img: NDArray[np.bool_], scale: nm, sigma: nm
) -> NDArray[np.float32]:
    """
    Pipe operation that smooth a binary image using a Gaussian kernel.

    Parameters
    ----------
    sigma : float
        Standard deviation of the Gaussian kernel.
    """
    if sigma == 0:
        return img.astype(np.float32)
    img = ~img
    dist: NDArray[np.float32] = ndi.distance_transform_edt(img)  # type: ignore
    blurred_mask = np.exp(-(dist**2) / 2 / (sigma / scale) ** 2, dtype=np.float32)
    return blurred_mask


def soft_otsu(sigma: nm = 1.0, radius: nm = 1.0, bins: int = 256):
    """
    Pipe operation of soft Otsu thresholding.

    This operation binarize an image using Otsu's method, dilate the edges and
    smooth the image using a Gaussian kernel.

    >>> from acryo.pipe import reader, soft_otsu
    >>> loader.align(
    ...     template=reader("path/to/template.mrc"),
    ...     mask=soft_otsu(2.0, 2.0),
    ... )

    Parameters
    ----------
    sigma : float, default is 1.0
        Standard deviation of the Gaussian kernel.
    radius : float, default is 1.0
        Radius of the structure element. If negative, erosion is applied.
    bins : int, default is 256
        Number of bins to build histogram.
    """
    out = gaussian_smooth(sigma) @ dilation(radius) @ threshold_otsu(bins)
    return out.with_name(f"soft_otsu({sigma=:.2f}, {radius=:.2f}, {bins=})")



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

import warnings
from typing import Callable, Sequence, TYPE_CHECKING, TypeVar
from functools import lru_cache, reduce, wraps

import numpy as np
from numpy.typing import NDArray
from dask import array as da
from scipy import ndimage as ndi
from scipy.spatial.transform import Rotation
from acryo._typed_scipy import fftn, rfftn, irfftn

if TYPE_CHECKING:
    from acryo._types import degree


def make_slice_and_pad(
    z0: int, z1: int, size: int
) -> tuple[slice, tuple[int, int], bool]:
    """
    Helper function for cropping images.

    This function calculates what slicing and padding are needed when an array
    is sliced by ``z0:z1``. Array must be padded when z0 is negative or z1 is
    outside the array size.
    """
    z0_pad = z1_pad = 0
    if z0 < 0:
        z0_pad = -z0
        z0 = 0
    elif size < z0:
        raise ValueError(f"Specified size is {size} but need to slice at {z0}:{z1}.")

    if size < z1:
        z1_pad = z1 - size
        z1 = size
    elif z1 < 0:
        raise ValueError(f"Specified size is {size} but need to slice at {z0}:{z1}.")

    out_of_bound = z0_pad != 0 or z1_pad != 0
    return slice(z0, z1), (z0_pad, z1_pad), out_of_bound


def compose_matrices(
    center: Sequence[float] | np.ndarray,
    rotators: Rotation | list[Rotation],
    output_center: Sequence[float] | np.ndarray | None = None,
) -> list[NDArray[np.float32]]:
    """Compose Affine matrices from an array shape and a Rotation object."""

    dz, dy, dx = center
    if output_center is None:
        output_center = center
    # center to corner
    translation_0 = np.array(
        [
            [1.0, 0.0, 0.0, dz],
            [0.0, 1.0, 0.0, dy],
            [0.0, 0.0, 1.0, dx],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype=np.float32,
    )
    # corner to center
    dz, dy, dx = output_center
    translation_1 = np.array(
        [
            [1.0, 0.0, 0.0, -dz],
            [0.0, 1.0, 0.0, -dy],
            [0.0, 0.0, 1.0, -dx],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype=np.float32,
    )

    matrices: list[np.ndarray] = []
    for rot in rotators:
        e_ = np.eye(4, dtype=np.float32)
        e_[:3, :3] = rot.as_matrix()
        matrices.append(translation_0 @ e_ @ translation_1)
    return matrices


def fourier_shell_correlation(
    img0: np.ndarray,
    img1: np.ndarray,
    dfreq: float = 0.02,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculate Fourier shell correlation for resolution analysis.

    Parameters
    ----------
    img0 : np.ndarray
        First input image.
    img1 : np.ndarray
        Second input image.
    dfreq : float, default is 0.02
        Difference of sampling frequency.

    Returns
    -------
    np.ndarray and np.ndarray
        Frequency and FSC.
    """
    shape = img0.shape

    freqs = np.meshgrid(
        *[np.fft.fftshift(np.fft.fftfreq(s, d=1.0)) for s in shape], indexing="ij"
    )

    r: np.ndarray = np.sqrt(sum(f**2 for f in freqs))

    # make radially separated labels
    labels = (r / dfreq).astype(np.uint16)
    nlabels = labels.max()

    out = np.empty(nlabels, dtype=np.float32)

    def radial_sum(arr: NDArray[np.float32]) -> NDArray[np.float32]:
        arr = np.asarray(arr)
        return ndi.sum_labels(arr, labels=labels, index=np.arange(0, nlabels))

    f0 = np.fft.fftshift(fftn(img0))
    f1 = np.fft.fftshift(fftn(img1))

    cov = f0.real * f1.real + f0.imag * f1.imag
    pw0 = f0.real**2 + f0.imag**2
    pw1 = f1.real**2 + f1.imag**2
    out = radial_sum(cov) / np.sqrt(radial_sum(pw0) * radial_sum(pw1))
    freq = (np.arange(len(out)) + 0.5) * dfreq
    return freq, out


def bin_image(img: np.ndarray | da.Array, binsize: int):
    """Bin an image."""
    _slices: list[slice] = []
    _shapes: list[int] = []
    for s in img.shape:
        npix, res = divmod(s, binsize)
        _slices.append(slice(None, s - res))
        _shapes.extend([npix, binsize])
    slices = tuple(_slices)
    shapes = tuple(_shapes)
    img_reshaped = img[slices].reshape(shapes)
    axis = tuple(i * 2 + 1 for i in range(img.ndim))
    return img_reshaped.sum(axis=axis)


def prepare_affine(
    img: da.Array,
    center: Sequence[float],
    output_shape: Sequence[int],
    rot: Rotation,
    order: int = 1,
) -> tuple[da.Array, NDArray[np.float32]]:
    output_center = np.array(output_shape) / 2 - 0.5
    slices: list[slice] = []
    pads: list[tuple[int, ...]] = []
    new_center: list[float] = []
    need_pad = False
    for c, s, s0 in zip(center, output_shape, img.shape):
        x0 = int(c - s / 2 - order)
        x1 = int(x0 + s + 2 * order + 1)
        _sl, _pad, _need_pad = make_slice_and_pad(x0, x1, s0)
        slices.append(_sl)
        pads.append(_pad)
        new_center.append(c - x0)
        need_pad = need_pad or _need_pad

    img0 = img[tuple(slices)]
    if need_pad:
        input = da.pad(img0, pads, mode="mean")
    else:
        input = img0
    mtx = compose_matrices(new_center, [rot], output_center=output_center)[0]
    return input, mtx


def prepare_affine_cornersafe(
    img: da.Array,
    center: Sequence[float],
    output_shape: Sequence[int],
    rot: Rotation,
    order: int = 1,
) -> tuple[da.Array, NDArray[np.float32]]:
    max_len = np.sqrt(np.sum(np.asarray(output_shape, dtype=np.float32) ** 2))
    output_center = np.array(output_shape) / 2 - 0.5
    half_len = max_len / 2
    slices: list[slice] = []
    pads: list[tuple[int, ...]] = []
    new_center: list[float] = []
    need_pad = False
    for c, s0 in zip(center, img.shape):
        x0 = int(c - half_len - order)
        x1 = int(x0 + max_len + 2 * order + 1)
        _sl, _pad, _need_pad = make_slice_and_pad(x0, x1, s0)
        slices.append(_sl)
        pads.append(_pad)
        new_center.append(c - x0)
        need_pad = need_pad or _need_pad

    img0 = img[tuple(slices)]
    if need_pad:
        input = da.pad(img0, pads, mode="mean")
    else:
        input = img0
    mtx = compose_matrices(new_center, [rot], output_center=output_center)[0]
    return input, mtx


def missing_wedge_mask(
    rotator: Rotation,
    tilt_range: tuple[degree, degree],
    shape: tuple[int, int, int],
) -> NDArray[np.float32]:
    """
    Create a binary mask that covers tomographical missing wedge.

    Note that the mask is not shifted to the center of the Fourier domain.
    ``np.fft.fftn(img) * mask`` will be the correct way to apply the mask.

    Parameters
    ----------
    rotator : Rotation
        The rotation object that describes the direction of the mask.
    tilt_range : tuple of float
        The range of tilt angles in degrees.
    shape : tuple of int
        The shape of the mask.

    Returns
    -------
    np.ndarray or float
        Missing wedge mask. If ``tilt_range`` is None, 1 will be returned.
    """
    normal0, normal1 = _get_unrotated_normals(tilt_range)
    shape_vector = np.array(shape, dtype=np.float32)
    rotator_inv = rotator.inv()
    normal0 = rotator_inv.apply(normal0 * shape_vector)
    normal1 = rotator_inv.apply(normal1 * shape_vector)
    vectors = _get_indices(shape)
    dot0 = vectors.dot(normal0)
    dot1 = vectors.dot(normal1)
    missing = dot0 * dot1 <= 0
    return missing


@lru_cache(maxsize=8)
def _get_unrotated_normals(
    tilt_range: tuple[degree, degree]
) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
    radmin, radmax = np.deg2rad(tilt_range)
    ang0 = np.pi - radmin
    ang1 = np.pi - radmax
    return (
        np.array([np.cos(ang0), 0, np.sin(ang0)], dtype=np.float32),
        np.array([np.cos(ang1), 0, np.sin(ang1)], dtype=np.float32),
    )


@lru_cache(maxsize=32)
def _get_indices(shape: tuple[int, ...]) -> NDArray[np.float32]:
    inds = np.indices(shape, dtype=np.float32)
    for ind, s in zip(inds, shape):
        # Note that the shifts in indices must resemble the shifts in fftshift.
        ind -= np.ceil(s / 2)
    return np.fft.fftshift(np.stack(list(inds), axis=-1), axes=(0, 1, 2))


# lowpass filter


def lowpass_filter_ft(
    img: NDArray[np.float32], cutoff: float, order: int = 2
) -> NDArray[np.complex64]:
    """Apply a low-pass filter and return the result in Fourier space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return fftn(img)
    weight = nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=False,
    )
    return weight * fftn(img)


def lowpass_filter(
    img: NDArray[np.float32], cutoff: float, order: int = 2
) -> NDArray[np.float32]:
    """Apply a low-pass filter and return the result in real space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return img
    weight = nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=True,
    )
    out = irfftn(weight * rfftn(img))
    return out.real


def highpass_filter_ft(
    img: NDArray[np.float32], cutoff: float, order: int = 2
) -> NDArray[np.complex64]:
    """Apply a high-pass filter and return the result in Fourier space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return np.zeros_like(img)
    weight = 1 - nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=False,
    )
    return weight * fftn(img)


def highpass_filter(
    img: NDArray[np.float32], cutoff: float, order: int = 2
) -> NDArray[np.float32]:
    """Apply a high-pass filter and return the result in real space."""
    if cutoff >= 0.5 * np.sqrt(img.ndim) or cutoff <= 0:
        return np.zeros_like(img)
    weight = 1 - nd_butterworth_weight(
        img.shape,
        cutoff,
        order,
        real=True,
    )
    out = irfftn(weight * rfftn(img))
    return out.real


# Modified from skimage.filters._fft_based
@lru_cache(maxsize=4)
def nd_butterworth_weight(
    shape: tuple[int, ...],
    cutoff: float,
    order: int,
    real: bool,
) -> NDArray[np.float32]:
    ranges = []
    for d in shape:
        axis = np.arange(-(d - 1) // 2, (d - 1) // 2 + 1, dtype=np.float32) / (
            d * cutoff
        )
        ranges.append(np.fft.ifftshift(axis**2))
    if real:
        limit = shape[-1] // 2 + 1
        ranges[-1] = ranges[-1][:limit]
    q2 = reduce(np.add, np.meshgrid(*ranges, indexing="ij", sparse=True))
    wfilt = 1 / (1 + q2**order)
    return wfilt


_F = TypeVar("_F", bound=Callable)


def deprecated_kwarg(old: str, new: str) -> Callable[[_F], _F]:
    def _inner(f):
        @wraps(f)
        def _func(*args, **kwargs):
            if old in kwargs:
                warnings.warn(
                    f"`{old}` is deprecated, use `{new}` instead",
                    DeprecationWarning,
                )
                kwargs[new] = kwargs.pop(old)
            return f(*args, **kwargs)

        return _func

    return _inner  # type: ignore



from __future__ import annotations
from typing import (
    Hashable,
    TYPE_CHECKING,
    Iterator,
    TypeVar,
    Generic,
)
import polars as pl

if TYPE_CHECKING:
    from polars.dataframe.groupby import GroupBy
    from .core import Molecules

_K = TypeVar("_K", bound=Hashable)


class MoleculeGroup(Generic[_K]):
    """A groupby-like object for molecules."""

    def __init__(self, group: GroupBy[pl.DataFrame]):
        self._group = group

    def __iter__(self) -> Iterator[tuple[_K, Molecules]]:
        from .core import Molecules

        for key, df in self._group:
            mole = Molecules.from_dataframe(df)
            yield key, mole  # type: ignore

    @property
    def features(self) -> GroupBy:
        """Return the groupby object."""
        return self._group



from ._masking import gaussian_smooth, soft_otsu, threshold_otsu, dilation
from ._transform import (
    center_by_mass,
    gaussian_filter,
    lowpass_filter,
    highpass_filter,
)
from ._curry import provider_function, converter_function
from ._imread import from_file, from_gaussian, from_array, from_atoms

__all__ = [
    "gaussian_smooth",
    "soft_otsu",
    "threshold_otsu",
    "dilation",
    "center_by_mass",
    "gaussian_filter",
    "lowpass_filter",
    "highpass_filter",
    "from_file",
    "from_gaussian",
    "from_array",
    "from_atoms",
    "provider_function",
    "converter_function",
]
